#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

""" Various Things needed for capability service"""

import inspect
from abc import ABC, abstractmethod
from typing import List, NamedTuple

# pylint: disable=C0116, C0301, R0903

from workspaces.system.schema import AbstractFile

CapabilityName = str


class QueueReport(NamedTuple):
    """
    I'm a report about what queues exist, how many executions are waiting in the queue,
    how many executions are executing in the queue, and how many concurrent executions
    are supported.
    """

    capability: str
    waiting: int
    executing: int
    concurrency_allowed: int


class QueueReporterIF(ABC):
    """
    An interface that defines a few report queries that are needed by the queue manager.
    """

    @abstractmethod
    def report(self) -> List[QueueReport]:
        """
        Retrieve a fresh report about what queues are in what states.
        """
        pass

    @abstractmethod
    def waiting_executions(self, queue: str, count: int) -> List["CapabilityExecution"]:
        """
        Retrieve the next `count` executions waiting in the named queue.
        :param queue:  the queue to check
        :param count:  the number to return
        :return:       the highest-priority CapabilityExecution
        """


class ExecutionManagerIF(ABC):
    """Manages capability execution system"""

    @abstractmethod
    def execute(self, execution: "CapabilityExecution"):
        """
        Start executing some execution.

        :param execution: the execution to start running
        """

    @abstractmethod
    def resume_execution(self, execution: "CapabilityExecution"):
        """
        Continue executing some execution.

        :param execution: the execution to start running
        """

    @abstractmethod
    def queue_execution(self, execution: "CapabilityExecution"):
        """
        Continue executing some execution.

        :param execution: the execution to start running
        """

    @abstractmethod
    def submit_workflow_request(
        self,
        execution_id: int,
        workflow_name: str,
        workflow_args: dict,
        files: List[AbstractFile],
    ):
        """
        Launch a workflow request.

        :param execution_id:   the execution ID to use
        :param workflow_name:  the name of the workflow to run
        :param workflow_args:  the arguments to the workflow (JSON dictionary)
        :param files:          files to submit
        """


class CapabilityQueueManagerIF(ABC):
    """
    The CapabilityQueueManager ensures that the right number of things are executing.
    If more things arrive, the queue manager must be notified. If some things finish, the
    queue manager must be notified. The queue manager ensures that the proper actions are
    taken in response to these notifications.
    """

    def execution_requested(self):
        """
        Notifies the queue manager that something new is requesting execution.
        """

    def execution_complete(self):
        """
        Notifies the queue manager that something has finished executing.
        """

    def reconfigure(self):
        """
        Notifies the queue manager that the system has been reconfigured (such as queue sizes adjusted).
        """


class CapabilityQueueIF(ABC):
    """
    Organizes capability executions in a priority order and makes it possible to control
    the number of concurrent executions or pause execution altogether
    """

    @abstractmethod
    def enqueue(self, execution: "CapabilityExecution", priority: int):
        raise NotImplementedError(f"{self.__class__.__name__}.{inspect.stack()[0][3]}")

    @abstractmethod
    def dequeue(self):
        raise NotImplementedError(f"{self.__class__.__name__}.{inspect.stack()[0][3]}")

    @abstractmethod
    def pause(self):
        raise NotImplementedError(f"{self.__class__.__name__}.{inspect.stack()[0][3]}")


class CapabilityServiceIF(ABC):
    """
    The capability service: clients access this to request capability runs.
    """
